//////Support_ExpandedAltFire//////
//By piber20
//BL_ID 3300
//http://piber20.com/

//////Documentation//////
//In short, this does two things:
//1. Makes doing things like recoil, spread, and multiple projectiles easier and cleaner. (from the weapon script's perspective)
//2. Adds support for an "alt fire" triggered by right click.
//This allows for something like a recreation of the shotgun from Half-Life, complete with firing multiple shells and the ability to "fire both barrels" by pressing right click, with minimal code required.
//Place this in your weapon mod's files and execute it.
//
//All below datablock values are to be used in the iten's image datablock.
//
//Standard fire image datablock values:
//projectileSpread = Defines the spread of fired projectiles.
//projectileCount = Defines the amount of projectiles the gun fires.
//playerRecoil = How hard the player is pushed back
//
//Alt fire image datablock values:
//hasAltShot = Set to true to enable the alt fire functionality. (right click)
//altProjectile = Projectile to fire on right click. Leave blank to use the item's normal left click projectile.
//altProjectileSpread = The spread of right click projectiles.
//altProjectileCount = Amount of right click projectiles.
//playerAltRecoil = How hard the player is pushed back from right click.
//altFireSound = Sound to play on right click.
//minAltShotTime = Delay between shots.
//
//Extra projectile datablock values:
//headshotDamage = Amount of damage to deal to the player if it was a headshot
//headshotMultiplier = Multiply damage by the number provided if it was a headshot. If you have provided a headshotDamage value, it'll multiply that instead of directDamage.
//headshotDamageType = DamageType of headshots. Will default to normal damage's damagetype if none is provided
//
//You can also use Support_RaycastingWeapons.cs features for the alt fire: (requires the script)
//raycastAltWeaponRange = Range of weapon (> 0)
//raycastAltWeaponTargets = Typemasks
//raycastAltWeaponPierceTargets = Typemasks (weapon fires through these targets, hitting each)
//raycastAltDirectDamage = Direct Damage
//raycastAltDirectDamageType = Damage Type ID
//raycastAltCritDirectDamageType = Critical Hit Damage Type ID
//raycastAltExplosionProjectile = Creates this projectile on impact
//raycastAltExplosionSound = AudioProfile of sound to play on impact
//raycastAltExplosionPlayerSound = AudioProfile of sound to play on hitting a player
//raycastAltExplosionBrickSound = AudioProfile of sound to play on hitting a brick
//raycastAltExplosionCritSound = AudioProfile of sound to play a critical impact
//raycastAltSpreadAmt = Spread radius of weapon
//raycastAltSpreadCount = Number of spreading projectiles
//raycastAltTracerProjectile = Fires tracer projectiles along the spread path
//raycastAltCritTracerProjectile = Fires tracer projectiles along the spread path if critical hit
//raycastAltFromMuzzle = Fire from muzzle point to muzzle vector instead of eye
//Though keep in mind, this calls onFire using these alt fire vars.
//
//You will also want to use the onAltFire function in your scripts. Here's what a basic setup looks like:
//function GunImage::onAltFire(%this, %obj, %slot)
//{
//	%obj.playThread(2, shiftAway);
//	Parent::onAltFire(%this, %obj, %slot);
//}
//It's very similar to onFire. You can use this hook-in to expand the functionality of your items via scripting.

if($Support::ExpandedAltFire::Version >= 5)
	return; //to stop repeat executions and so earlier versions wont overwrite newer ones

$Support::ExpandedAltFire::Version = 5;

package SupportExpandedAltFirePackage
{
	function WeaponImage::onFire(%this, %obj, %slot)
	{
		if((%obj.lastExpandedFireTime + %this.minShotTime) > getSimTime())
			return Parent::onFire(%this, %obj, %slot);
		
		%obj.lastExpandedFireTime = getSimTime();
		
		//recoil
		if(mfloor(%this.playerRecoil) != 0)
		{
			%recoil = %this.playerRecoil;
			if(%recoil > 0)
				%recoil = %recoil - %recoil - %recoil;
			
			%fvec = %obj.getForwardVector();
			%fX = getWord(%fvec,0);
			%fY = getWord(%fvec,1);
			
			%evec = %obj.getEyeVector();
			%eX = getWord(%evec,0);
			%eY = getWord(%evec,1);
			%eZ = getWord(%evec,2);
			
			%eXY = mSqrt(%eX*%eX+%eY*%eY);
			
			%aimVec = %fX*%eXY SPC %fY*%eXY SPC %eZ;

			%obj.setVelocity(VectorAdd(%obj.getVelocity(), VectorScale(%aimVec, %recoil)));
		}

		//projectiles
		if(%this.projectileSpread !$= "" || %this.projectileCount !$= "")
		{
			if(%this.raycastWeaponRange > 0) //they're using expandedaltfire main fire stuff with a weapon that has raycasts. print an error and return to onFire so we dont break raycasts
			{
				if(!%this.hideExpandedAltFireRaycastError) //if they really wanted to, they could preemptively set this. would be kinda lame tho
				{
					error("ERROR: projectileSpread or projectileCount detected on " @ %this.getName() @ " which uses raycasts, use raycastSpreadAmt or raycastSpreadCount instead.");
					%this.hideExpandedAltFireRaycastError = true; //only print the error once, we dont want to spam the console
				}
				
				%this.raycastSpreadAmt = %this.projectileSpread;
				%this.raycastSpreadCount = %this.projectileCount;
				
				return Parent::onFire(%this, %obj, %slot);
			}
			
			%projectile = %this.projectile;
			%spread = mAbs(%this.projectileSpread);
			%amount = mfloor(%this.projectileCount);
			if(%amount <= 0)
				%amount = 1;

			for(%shell = 0; %shell < %amount; %shell++)
			{
				%vector = %obj.getMuzzleVector(%slot);
				%objectVelocity = %obj.getVelocity();
				%vector1 = VectorScale(%vector, %projectile.muzzleVelocity);
				%vector2 = VectorScale(%objectVelocity, %projectile.velInheritFactor);
				%velocity = VectorAdd(%vector1,%vector2);
				%x = (getRandom() - 0.5) * 10 * 3.1415926 * %spread;
				%y = (getRandom() - 0.5) * 10 * 3.1415926 * %spread;
				%z = (getRandom() - 0.5) * 10 * 3.1415926 * %spread;
				%mat = MatrixCreateFromEuler(%x @ " " @ %y @ " " @ %z);
				%velocity = MatrixMulVector(%mat, %velocity);

				%p = new (%this.projectileType)()
				{
					dataBlock = %projectile;
					initialVelocity = %velocity;
					initialPosition = %obj.getMuzzlePoint(%slot);
					sourceObject = %obj;
					sourceSlot = %slot;
					client = %obj.client;
				};
				MissionCleanup.add(%p);
			}
		}
		else
			return Parent::onFire(%this, %obj, %slot); //very very important we have this here, it allows default and non-expandedaltfire weapons to work
	}
	
	function Armor::onTrigger(%this, %player, %trigger, %val)
	{
		Parent::onTrigger(%this, %player, %trigger, %val);
		
		if(%trigger == 4 && %val) //4 == right click
		{
			//get what they're holding
			%image = %player.getMountedImage(0);
			%image1 = %player.getMountedImage(1);
			
			//does it exist?
			if(!isObject(%image))
				%noItem++;
			if(!isObject(%image1))
				%noItem++;
			
			if(%noItem >= 2)
				return;
			
			//does it have an alt shot?
			if(!%image.hasAltShot)
				%noShot++;
			if(!%image1.hasAltShot)
				%noShot++;
			
			if(%noShot >= 2)
				return;
			
			//fire the correct guns
			if(isObject(%image) && %image.hasAltShot && (%player.lastAltFireTime + %image.minAltShotTime) <= getSimTime())
				%image.onAltFire(%player, 0);
			if(isObject(%image1) && %image1.hasAltShot && (%player.lastAltFireTime + %image1.minAltShotTime) <= getSimTime())
				%image1.onAltFire(%player, 1);
		}
	}
	
	function disconnect()
	{
		//reset version variable when the server ends
		$Support::ExpandedAltFire::Version = "";
		
		return parent::disconnect();
	}

	function ProjectileData::damage(%this, %obj, %col, %fade, %pos, %normal)
	{
		if(mceil(%this.headshotMultiplier) > 0 || mceil(%this.headshotDamage) > 0) //make sure any of the vars are defined, mceil for a blank string would be 0
		{
			if(%this.headshotMultiplier != 1 || %this.headshotDamage != %this.directDamage) //we have no reason to do this if the multiplier is 1 and the damage is the same as direct damage
			{
				if(%col.getType() & $TypeMasks::PlayerObjectType) //make sure we hit a player object
				{
					//base damage
					%baseDamage = %this.directDamage;
					if(%this.headshotDamage !$= "" && %this.headshotDamage > 0)
						%baseDamage = %this.headshotDamage;
					
					if(%baseDamage > 0)
					{
						%colscale = getWord(%col.getScale(), 2);
						if(getword(%pos, 2) > getword(%col.getWorldBoxCenter(), 2) - (3.3 * %colscale))
						{
							//multiplier
							%multiplier = 1.0;
							if(%this.headshotMultiplier !$= "" && %this.headshotMultiplier > 0)
								%multiplier = %this.headshotMultiplier;
							
							//math
							%damageAmount = %baseDamage * %multiplier;
							
							//damage type
							%damageType = $DamageType::Direct;
							if(%this.directDamageType !$= "")
								%damageType = %this.directDamageType;
							if(%this.headshotDamageType !$= "")
								%damageType = %this.headshotDamageType;

							//deal the damage
							%col.damage(%obj, %pos, %damageAmount, %damageType);
							return;
						}
					}
				}
			}
		}
		Parent::damage(%this, %obj, %col, %fade, %pos, %normal);
	}
};
activatePackage(SupportExpandedAltFirePackage);

function WeaponImage::onAltFire(%this, %obj, %slot) //for use with weapons that have right click functionality
{
	%obj.lastAltFireTime = getSimTime();
	
	//recoil
	if(mfloor(%this.playerAltRecoil) != 0)
	{
		%recoil = %this.playerAltRecoil;
		if(%recoil > 0)
			%recoil = %recoil - %recoil - %recoil;
		
		%fvec = %obj.getForwardVector();
		%fX = getWord(%fvec,0);
		%fY = getWord(%fvec,1);
		
		%evec = %obj.getEyeVector();
		%eX = getWord(%evec,0);
		%eY = getWord(%evec,1);
		%eZ = getWord(%evec,2);
		
		%eXY = mSqrt(%eX*%eX+%eY*%eY);
		
		%aimVec = %fX*%eXY SPC %fY*%eXY SPC %eZ;

		%obj.setVelocity(VectorAdd(%obj.getVelocity(), VectorScale(%aimVec, %recoil)));
	}

	//sound
	if(isObject(%this.altFireSound))
		serverPlay3D(%this.altFireSound, %obj.getPosition());
	
	//raycasting - we replace the main fire raycast vars with the alt fire ones and call onFire. hacky but it works
	if(%this.raycastAltWeaponRange > 0)
	{
		//of course, we're going to require the support mod so weapons can use this functionality
		if(mfloor($SpaceMods::Server::RaycastingWeaponsVersion) <= 0) //sets it to -1 when it disables itself
		{
			error("ERROR: Tried to do an alt fire raycast without Support_RaycastingWeapons.cs");
			return;
		}
		
		//store the original raycast vars so they're not permanently gone
		%oldWeaponRange = %this.raycastWeaponRange;
		%oldWeaponTargets = %this.raycastWeaponTargets;
		%oldWeaponPierceTargets = %this.raycastWeaponPierceTargets;
		%oldFromMuzzle = %this.raycastFromMuzzle;
		%oldSpreadCount = %this.raycastSpreadCount;
		%oldSpreadAmt = %this.raycastSpreadAmt;
		%oldTracerProjectile = %this.raycastTracerProjectile;
		%oldCritTracerProjectile = %this.raycastCritTracerProjectile;
		%oldDirectDamage = %this.raycastDirectDamage;
		%oldExplosionProjectile = %this.raycastExplosionProjectile;
		%oldExplosionSound = %this.raycastExplosionSound;
		%oldExplosionCritSound = %this.raycastExplosionCritSound;
		%oldExplosionPlayerSound = %this.raycastExplosionPlayerSound;
		%oldExplosionBrickSound = %this.raycastExplosionBrickSound;
		%oldDirectDamageType = %this.raycastDirectDamageType;
		%oldCritDirectDamageType = %this.raycastCritDirectDamageType;
		%oldImpactImpulse = %this.raycastImpactImpulse;
		%oldVerticalImpulse = %this.raycastVerticalImpulse;
		
		//replace the raycast vars with the raycastalt vars
		%this.raycastWeaponRange = %this.raycastAltWeaponRange;
		%this.raycastWeaponTargets = %this.raycastAltWeaponTargets;
		%this.raycastWeaponPierceTargets = %this.raycastAltWeaponPierceTargets;
		%this.raycastFromMuzzle = %this.raycastAltFromMuzzle;
		%this.raycastSpreadCount = %this.raycastAltSpreadCount;
		%this.raycastSpreadAmt = %this.raycastAltSpreadAmt;
		%this.raycastTracerProjectile = %this.raycastAltTracerProjectile;
		%this.raycastCritTracerProjectile = %this.raycastAltCritTracerProjectile;
		%this.raycastDirectDamage = %this.raycastAltDirectDamage;
		%this.raycastExplosionProjectile = %this.raycastAltExplosionProjectile;
		%this.raycastExplosionSound = %this.raycastAltExplosionSound;
		%this.raycastExplosionCritSound = %this.raycastAltExplosionCritSound;
		%this.raycastExplosionPlayerSound = %this.raycastAltExplosionPlayerSound;
		%this.raycastExplosionBrickSound = %this.raycastAltExplosionBrickSound;
		%this.raycastDirectDamageType = %this.raycastAltDirectDamageType;
		%this.raycastCritDirectDamageType = %this.raycastAltCritDirectDamageType;
		%this.raycastImpactImpulse = %this.raycastAltImpactImpulse;
		%this.raycastVerticalImpulse = %this.raycastAltVerticalImpulse;
		
		//call onFire using our replaced vars
		%this.onFire(%obj, %slot);
		
		//restore the original raycast vars
		%this.raycastWeaponRange = %oldWeaponRange;
		%this.raycastWeaponTargets = %oldWeaponTargets;
		%this.raycastWeaponPierceTargets = %oldWeaponPierceTargets;
		%this.raycastFromMuzzle = %oldFromMuzzle;
		%this.raycastSpreadCount = %oldSpreadCount;
		%this.raycastSpreadAmt = %oldSpreadAmt;
		%this.raycastTracerProjectile = %oldTracerProjectile;
		%this.raycastCritTracerProjectile = %oldCritTracerProjectile;
		%this.raycastDirectDamage = %oldDirectDamage;
		%this.raycastExplosionProjectile = %oldExplosionProjectile;
		%this.raycastExplosionSound = %oldExplosionSound;
		%this.raycastExplosionCritSound = %oldExplosionCritSound;
		%this.raycastExplosionPlayerSound = %oldExplosionPlayerSound;
		%this.raycastExplosionBrickSound = %oldExplosionBrickSound;
		%this.raycastDirectDamageType = %oldDirectDamageType;
		%this.raycastCritDirectDamageType = %oldCritDirectDamageType;
		%this.raycastImpactImpulse = %oldImpactImpulse;
		%this.raycastVerticalImpulse = %oldVerticalImpulse;
	}
	else //standard projectiles
	{
		if(%this.altProjectile $= "")
			%projectile = %this.projectile;
		else
			%projectile = %this.altProjectile;
		
		%spread = mAbs(%this.altProjectileSpread);
		
		%amount = mfloor(%this.altProjectileCount);
		if(%amount <= 0)
			%amount = 1;

		for(%shell = 0; %shell < %amount; %shell++)
		{
			%vector = %obj.getMuzzleVector(%slot);
			%objectVelocity = %obj.getVelocity();
			%vector1 = VectorScale(%vector, %projectile.muzzleVelocity);
			%vector2 = VectorScale(%objectVelocity, %projectile.velInheritFactor);
			%velocity = VectorAdd(%vector1,%vector2);
			%x = (getRandom() - 0.5) * 10 * 3.1415926 * %spread;
			%y = (getRandom() - 0.5) * 10 * 3.1415926 * %spread;
			%z = (getRandom() - 0.5) * 10 * 3.1415926 * %spread;
			%mat = MatrixCreateFromEuler(%x @ " " @ %y @ " " @ %z);
			%velocity = MatrixMulVector(%mat, %velocity);

			%p = new (%this.projectileType)()
			{
				dataBlock = %projectile;
				initialVelocity = %velocity;
				initialPosition = %obj.getMuzzlePoint(%slot);
				sourceObject = %obj;
				sourceSlot = %slot;
				client = %obj.client;
			};
			MissionCleanup.add(%p);
		}
	}
}